from syncloud.app import runner
from syncloud.app import logger

from util import copyfile, create_link, Settings

from os import makedirs
from os.path import exists, join

from keygen import KeyGen

BITS = 2048

class SshInitD:
    def __init__(self, filename):
        self.log = logger.get_logger('SshInitD')
        self.filename=filename
        copyfile('/etc/init.d/ssh', self.filename, self.log)
        self.__load()
        self.__replace('/etc/init.d/ssh', self.filename)

    def __load(self):
        with open(self.filename, 'r') as f:
            self.data = f.read()

    def __replace(self, old, new):
        self.data = self.data.replace(old, new)

    def pid_file(self, filename):
        self.__replace('/var/run/sshd.pid', filename)

    def running_folder(self, path):
        self.__replace('/var/run/sshd', path)

    def executable(self, filename):
        self.__replace('/usr/sbin/sshd', filename)

    def service_name(self, name):
        self.__replace(' sshd ', ' {0} '.format(name))
        self.__replace('"sshd"', '"{0}"'.format(name))
        self.__replace('"sshd"', '"{0}"'.format(name))
        self.__replace('Provides:		sshd', 'Provides:		{0}'.format(name))

    def default_config(self, filename):
        self.__replace('/etc/default/ssh', filename)

    def save(self):
        with open(self.filename, 'w') as f:
            f.write(self.data)


def update_defaults(name, log):
    runner.call('update-rc.d -f {0} defaults'.format(name), log, shell=True)

def add_to_initd(name, log):
    runner.call('insserv --verbose --remove {0}'.format(name), log, shell=True)
    runner.call('insserv --verbose --default {0}'.format(name), log, shell=True)
    runner.call('systemctl --system daemon-reload', log, shell=True)

def restart_service(name, log):
    runner.call('service {0} stop'.format(name), log, shell=True)
    runner.call('service {0} start'.format(name), log, shell=True)


class SshInitD:
    def __init__(self, filename):
        self.log = logger.get_logger('SshInitD')
        self.filename=filename
        copyfile('/etc/init.d/ssh', self.filename, self.log)
        self.__load()
        self.__replace('/etc/init.d/ssh', self.filename)

    def __load(self):
        with open(self.filename, 'r') as f:
            self.data = f.read()

    def __replace(self, old, new):
        self.data = self.data.replace(old, new)

    def pid_file(self, filename):
        self.__replace('/var/run/sshd.pid', filename)

    def running_folder(self, path):
        self.__replace('/var/run/sshd', path)

    def executable(self, filename):
        self.__replace('/usr/sbin/sshd', filename)

    def service_name(self, name):
        self.__replace(' sshd ', ' {0} '.format(name))
        self.__replace('"sshd"', '"{0}"'.format(name))
        self.__replace('"sshd"', '"{0}"'.format(name))
        self.__replace('Provides:		sshd', 'Provides:		{0}'.format(name))

    def default_config(self, filename):
        self.__replace('/etc/default/ssh', filename)

    def save(self):
        with open(self.filename, 'w') as f:
            f.write(self.data)


def yes_no(value):
    if value:
        return 'yes'
    else:
        return 'no'

class SshServer:
    def __init__(self):
        self.log = logger.get_logger('SshServer')
        self.key_gen = KeyGen()

    def get_server_keys(self):
        rsa_key = '/etc/ssh/ssh_syncloud_host_rsa_key'
        dsa_key = '/etc/ssh/ssh_syncloud_host_dsa_key'
        ecdsa_key = '/etc/ssh/ssh_syncloud_host_ecdsa_key'
        if not exists(rsa_key):
            self.key_gen.generate_into_file('rsa', rsa_key, BITS, overwrite=True)
        if not exists(dsa_key):
            self.key_gen.generate_into_file('dsa', dsa_key, overwrite=True)
        if not exists(ecdsa_key):
            self.key_gen.generate_into_file('ecdsa', ecdsa_key, overwrite=True)
        return (rsa_key, dsa_key, ecdsa_key)


    def setup(self, name, service_port, password_authentication):
        rsa_key, dsa_key, ecdsa_key = self.get_server_keys()

        ssh_name = 'ssh_{0}'.format(name)
        sshd_name = 'sshd_{0}'.format(name)

        ssh_config_filename = '/etc/ssh/{0}_config'.format(ssh_name)
        sshd_config_filename = '/etc/ssh/{0}_config'.format(sshd_name)

        copyfile('/etc/ssh/ssh_config', ssh_config_filename, self.log)

        sshd_pid_filename = '/var/run/sshd_{0}.pid'.format(name)

        copyfile('/etc/ssh/sshd_config', sshd_config_filename, self.log)
        sshd_config = Settings(sshd_config_filename, delimeter=' ')
        sshd_config.set('PasswordAuthentication', yes_no(password_authentication))
        sshd_config.set('PubkeyAuthentication', yes_no(not password_authentication))
        if not password_authentication:
            certificates_path = '/root/.ssh_{0}'.format(name)
            # certificates_path = '/root/.ssh'
            authorized_keys_filename = join(certificates_path, 'authorized_keys')
            sshd_config.set('AuthorizedKeysFile', authorized_keys_filename)
        sshd_config.set('Port', service_port)
        sshd_config.set('PidFile', sshd_pid_filename)
        sshd_config.set('HostKey', rsa_key, add_to_existing=False)
        sshd_config.set('HostKey', dsa_key, add_to_existing=True)
        sshd_config.set('HostKey', ecdsa_key, add_to_existing=True)
        sshd_config.save()

        sshd_bin_filename = '/usr/sbin/{0}'.format(sshd_name)
        sshd_pam_filename = '/etc/pam.d/{0}'.format(sshd_name)

        create_link('/usr/sbin/sshd', sshd_bin_filename, self.log)
        create_link('/etc/pam.d/sshd', sshd_pam_filename, self.log)

        ssh_default_filename = '/etc/default/{0}'.format(ssh_name)

        copyfile('/etc/default/ssh', ssh_default_filename, self.log)
        default_parameter = Settings(ssh_default_filename, delimeter='=')
        default_parameter.set('SSHD_OPTS', '"-f {0}"'.format(sshd_config_filename))
        default_parameter.save()
        update_defaults(ssh_name, self.log)

        sshd_working_folder = '/var/run/{0}'.format(sshd_name)
        if not exists(sshd_working_folder):
            makedirs(sshd_working_folder)

        ssh_initd_filename = '/etc/init.d/{0}'.format(ssh_name)
        ssh_init_d = SshInitD(ssh_initd_filename)
        ssh_init_d.pid_file(sshd_pid_filename)
        ssh_init_d.running_folder(sshd_working_folder)
        ssh_init_d.executable(sshd_bin_filename)
        ssh_init_d.service_name(sshd_name)
        ssh_init_d.default_config(ssh_default_filename)
        ssh_init_d.save()

        add_to_initd(ssh_name, self.log)
        restart_service(ssh_name, self.log)

        if not password_authentication:
            if not exists(certificates_path):
                makedirs(certificates_path)

            private, public = self.key_gen.generate('rsa', BITS)

            with open(authorized_keys_filename, 'a+') as keys_file:
                keys_file.write(public)
                keys_file.write('\n')

            private_key_filename = join(certificates_path, 'id_rsa_syncloud_master')
            with open(private_key_filename, 'w+') as private_key_file:
                private_key_file.write(private)

            return private
